# ----------------------------------------------------------------------------------
# Electrum plugin for the Digital Bitbox hardware wallet by Shift Devices AG
# digitalbitbox.com
#

import base64
import binascii
import hashlib
import hmac
import json
import math
import os
import re
import struct
import sys
import time
import copy
from typing import TYPE_CHECKING, Optional

import electrum_ecc as ecc

from electrum.crypto import sha256d, EncodeAES_bytes, DecodeAES_bytes, hmac_oneshot
from electrum.bitcoin import public_key_to_p2pkh, usermessage_magic, verify_usermessage_with_address
from electrum.bip32 import BIP32Node, convert_bip32_intpath_to_strpath, is_all_public_derivation
from electrum.bip32 import normalize_bip32_derivation
from electrum import descriptor
from electrum.wallet import Standard_Wallet
from electrum import constants
from electrum.transaction import Transaction, PartialTransaction, PartialTxInput, Sighash
from electrum.i18n import _
from electrum.keystore import Hardware_KeyStore
from electrum.util import to_string, UserCancelled, UserFacingException, bfh, ChoiceItem
from electrum.network import Network
from electrum.logging import get_logger
from electrum.plugin import runs_in_hwd_thread, run_in_hwd_thread

from electrum.hw_wallet import HW_PluginBase, HardwareClientBase, HardwareHandlerBase
from electrum.hw_wallet.plugin import OperationCancelled

if TYPE_CHECKING:
    from electrum.plugin import DeviceInfo
    from electrum.wizard import NewWalletWizard

_logger = get_logger(__name__)


try:
    import hid
    DIGIBOX = True
except ImportError as e:
    DIGIBOX = False


class DeviceErased(UserFacingException):
    pass

# ----------------------------------------------------------------------------------
# USB HID interface
#


def to_hexstr(s):
    return binascii.hexlify(s).decode('ascii')


def derive_keys(x):
    h = sha256d(x)
    h = hashlib.sha512(h).digest()
    return (h[:32],h[32:])


MIN_MAJOR_VERSION = 5

ENCRYPTION_PRIVKEY_KEY = 'encryptionprivkey'
CHANNEL_ID_KEY = 'comserverchannelid'


class DigitalBitbox_Client(HardwareClientBase):
    def __init__(self, plugin, hidDevice):
        HardwareClientBase.__init__(self, plugin=plugin)
        self.dbb_hid = hidDevice
        self.opened = True
        self.password = None
        self.isInitialized = False
        self.setupRunning = False
        self.usbReportSize = 64  # firmware > v2.0.0

    def device_model_name(self) -> Optional[str]:
        return 'Digital BitBox'

    @runs_in_hwd_thread
    def close(self):
        if self.opened:
            try:
                self.dbb_hid.close()
            except Exception:
                pass
        self.opened = False

    def is_pairable(self):
        return True

    def is_initialized(self):
        return self.dbb_has_password()

    def is_paired(self):
        return self.password is not None

    def has_usable_connection_with_device(self):
        try:
            self.dbb_has_password()
        except BaseException:
            return False
        return True

    def _get_xpub(self, bip32_path: str):
        bip32_path = normalize_bip32_derivation(bip32_path, hardened_char="'")
        if self.check_device_dialog():
            return self.hid_send_encrypt(('{"xpub": "%s"}' % bip32_path).encode('utf8'))

    def get_xpub(self, bip32_path, xtype):
        assert xtype in self.plugin.SUPPORTED_XTYPES

        if is_all_public_derivation(bip32_path):
            raise UserFacingException(_('This device does not reveal xpubs corresponding to non-hardened paths'))

        reply = self._get_xpub(bip32_path)
        if reply:
            xpub = reply['xpub']
            # Change type of xpub to the requested type. The firmware
            # only ever returns the mainnet standard type, but it is agnostic
            # to the type when signing.
            if xtype != 'standard' or constants.net.TESTNET:
                node = BIP32Node.from_xkey(xpub, net=constants.BitcoinMainnet)
                xpub = node._replace(xtype=xtype).to_xpub()
            return xpub
        else:
            raise Exception('no reply')

    def get_soft_device_id(self):
        return None

    def dbb_has_password(self):
        reply = self.hid_send_plain(b'{"ping":""}')
        if 'ping' not in reply:
            raise UserFacingException(_('Device communication error. Please unplug and replug your Digital Bitbox.'))
        if reply['ping'] == 'password':
            return True
        return False

    def stretch_key(self, key: bytes):
        return to_hexstr(hashlib.pbkdf2_hmac('sha512', key, b'Digital Bitbox', iterations = 20480))

    def backup_password_dialog(self):
        msg = _("Enter the password used when the backup was created:")
        while True:
            password = self.handler.get_passphrase(msg, False)
            if password is None:
                return None
            if len(password) < 4:
                msg = _("Password must have at least 4 characters.") \
                      + "\n\n" + _("Enter password:")
            elif len(password) > 64:
                msg = _("Password must have less than 64 characters.") \
                      + "\n\n" + _("Enter password:")
            else:
                return password.encode('utf8')

    def password_dialog(self, msg):
        while True:
            password = self.handler.get_passphrase(msg, False)
            if password is None:
                return False
            if len(password) < 4:
                msg = _("Password must have at least 4 characters.") + \
                      "\n\n" + _("Enter password:")
            elif len(password) > 64:
                msg = _("Password must have less than 64 characters.") + \
                      "\n\n" + _("Enter password:")
            else:
                self.password = password.encode('utf8')
                return True

    def check_firmware_version(self):
        match = re.search(r'v([0-9])+\.[0-9]+\.[0-9]+',
                          run_in_hwd_thread(self.dbb_hid.get_serial_number_string))
        if match is None:
            raise Exception("error detecting firmware version")
        major_version = int(match.group(1))
        if major_version < MIN_MAJOR_VERSION:
            raise Exception("Please upgrade to the newest firmware using the BitBox Desktop app: https://shiftcrypto.ch/start")

    def check_device_dialog(self):
        self.check_firmware_version()
        # Set password if fresh device
        if self.password is None and not self.dbb_has_password():
            if not self.setupRunning:
                return False # A fresh device cannot connect to an existing wallet
            msg = _("An uninitialized Digital Bitbox is detected.") + " " + \
                  _("Enter a new password below.") + "\n\n" + \
                  _("REMEMBER THE PASSWORD!") + "\n\n" + \
                  _("You cannot access your coins or a backup without the password.") + "\n" + \
                  _("A backup is saved automatically when generating a new wallet.")
            if self.password_dialog(msg):
                reply = self.hid_send_plain(b'{"password":"' + self.password + b'"}')
            else:
                return False

        # Get password from user if not yet set
        msg = _("Enter your Digital Bitbox password:")
        while self.password is None:
            if not self.password_dialog(msg):
                raise UserCancelled()
            reply = self.hid_send_encrypt(b'{"led":"blink"}')
            if 'error' in reply:
                self.password = None
                if reply['error']['code'] == 109:
                    msg = _("Incorrect password entered.") + "\n\n" + \
                          reply['error']['message'] + "\n\n" + \
                          _("Enter your Digital Bitbox password:")
                else:
                    # Should never occur
                    msg = _("Unexpected error occurred.") + "\n\n" + \
                          reply['error']['message'] + "\n\n" + \
                          _("Enter your Digital Bitbox password:")

        # Initialize device if not yet initialized
        if not self.setupRunning:
            self.isInitialized = True # Wallet exists. Electrum code later checks if the device matches the wallet
        elif not self.isInitialized:
            reply = self.hid_send_encrypt(b'{"device":"info"}')
            if reply['device']['id'] != "":
                self.recover_or_erase_dialog() # Already seeded
            else:
                self.seed_device_dialog() # Seed if not initialized
            self.mobile_pairing_dialog()
        return self.isInitialized

    def recover_or_erase_dialog(self):
        msg = _("The Digital Bitbox is already seeded. Choose an option:") + "\n"
        choices = [
            ChoiceItem(key="create", label=_("Create a wallet using the current seed")),
            ChoiceItem(key="erase", label=_("Erase the Digital Bitbox")),
        ]
        reply = self.handler.query_choice(msg, choices)
        if reply is None:
            raise UserCancelled()
        if reply == "erase":
            self.dbb_erase()
        else:
            if self.hid_send_encrypt(b'{"device":"info"}')['device']['lock']:
                raise UserFacingException(_("Full 2FA enabled. This is not supported yet."))
            # Use existing seed
        self.isInitialized = True

    def seed_device_dialog(self):
        msg = _("Choose how to initialize your Digital Bitbox:") + "\n"
        choices = [
            ChoiceItem(key="generate", label=_("Generate a new random wallet")),
            ChoiceItem(key="load", label=_("Load a wallet from the micro SD card")),
        ]
        reply = self.handler.query_choice(msg, choices)
        if reply is None:
            raise UserCancelled()
        if reply == "generate":
            self.dbb_generate_wallet()
        else:
            if not self.dbb_load_backup(show_msg=False):
                return
        self.isInitialized = True

    def mobile_pairing_dialog(self):
        dbb_user_dir = None
        if sys.platform == 'darwin':
            dbb_user_dir = os.path.join(os.environ.get("HOME", ""), "Library", "Application Support", "DBB")
        elif sys.platform == 'win32':
            dbb_user_dir = os.path.join(os.environ["APPDATA"], "DBB")
        else:
            dbb_user_dir = os.path.join(os.environ["HOME"], ".dbb")

        if not dbb_user_dir:
            return

        try:
            with open(os.path.join(dbb_user_dir, "config.dat")) as f:
                dbb_config = json.load(f)
        except (FileNotFoundError, json.JSONDecodeError):
            return

        if ENCRYPTION_PRIVKEY_KEY not in dbb_config or CHANNEL_ID_KEY not in dbb_config:
            return

        choices = [
            ChoiceItem(key=0, label=_('Do not pair')),
            ChoiceItem(key=1, label=_('Import pairing from the Digital Bitbox desktop app')),
        ]
        reply = self.handler.query_choice(_('Mobile pairing options'), choices)
        if reply is None:
            raise UserCancelled()

        if reply == 0:
            if self.plugin.is_mobile_paired():
                del self.plugin.digitalbitbox_config[ENCRYPTION_PRIVKEY_KEY]
                del self.plugin.digitalbitbox_config[CHANNEL_ID_KEY]
        elif reply == 1:
            # import pairing from dbb app
            self.plugin.digitalbitbox_config[ENCRYPTION_PRIVKEY_KEY] = dbb_config[ENCRYPTION_PRIVKEY_KEY]
            self.plugin.digitalbitbox_config[CHANNEL_ID_KEY] = dbb_config[CHANNEL_ID_KEY]
        self.plugin.config.set_key('digitalbitbox', self.plugin.digitalbitbox_config)

    def dbb_generate_wallet(self):
        key = self.stretch_key(self.password)
        filename = ("Electrum-" + time.strftime("%Y-%m-%d-%H-%M-%S") + ".pdf")
        msg = ('{"seed":{"source": "create", "key": "%s", "filename": "%s", "entropy": "%s"}}' % (key, filename, to_hexstr(os.urandom(32)))).encode('utf8')
        reply = self.hid_send_encrypt(msg)
        if 'error' in reply:
            raise UserFacingException(reply['error']['message'])

    def dbb_erase(self):
        self.handler.show_message(_("Are you sure you want to erase the Digital Bitbox?") + "\n\n" +
                                  _("To continue, touch the Digital Bitbox's light for 3 seconds.") + "\n\n" +
                                  _("To cancel, briefly touch the light or wait for the timeout."))
        hid_reply = self.hid_send_encrypt(b'{"reset":"__ERASE__"}')
        self.handler.finished()
        if 'error' in hid_reply:
            if hid_reply['error'].get('code') in (600, 601):
                raise OperationCancelled()
            raise UserFacingException(hid_reply['error']['message'])
        else:
            self.password = None
            raise DeviceErased('Device erased')

    def dbb_load_backup(self, show_msg=True):
        backups = self.hid_send_encrypt(b'{"backup":"list"}')
        if 'error' in backups:
            raise UserFacingException(backups['error']['message'])
        backup_choices = [ChoiceItem(key=idx, label=v) for (idx, v) in enumerate(backups['backup'])]
        f = self.handler.query_choice(_("Choose a backup file:"), backup_choices)
        if f is None:
            raise UserCancelled()
        key = self.backup_password_dialog()
        if key is None:
            raise UserCancelled('No backup password provided')
        key = self.stretch_key(key)
        if show_msg:
            self.handler.show_message(_("Loading backup...") + "\n\n" +
                                      _("To continue, touch the Digital Bitbox's light for 3 seconds.") + "\n\n" +
                                      _("To cancel, briefly touch the light or wait for the timeout."))
        msg = ('{"seed":{"source": "backup", "key": "%s", "filename": "%s"}}' % (key, backups['backup'][f])).encode('utf8')
        hid_reply = self.hid_send_encrypt(msg)
        self.handler.finished()
        if 'error' in hid_reply:
            if hid_reply['error'].get('code') in (600, 601):
                raise OperationCancelled()
            raise UserFacingException(hid_reply['error']['message'])
        return True

    @runs_in_hwd_thread
    def hid_send_frame(self, data):
        HWW_CID = 0xFF000000
        HWW_CMD = 0x80 + 0x40 + 0x01
        data_len = len(data)
        seq = 0
        idx = 0
        write = []
        while idx < data_len:
            if idx == 0:
                # INIT frame
                write = data[idx : idx + min(data_len, self.usbReportSize - 7)]
                self.dbb_hid.write(b'\0' + struct.pack(">IBH", HWW_CID, HWW_CMD, data_len & 0xFFFF) + write + b'\xEE' * (self.usbReportSize - 7 - len(write)))
            else:
                # CONT frame
                write = data[idx : idx + min(data_len, self.usbReportSize - 5)]
                self.dbb_hid.write(b'\0' + struct.pack(">IB", HWW_CID, seq) + write + b'\xEE' * (self.usbReportSize - 5 - len(write)))
                seq += 1
            idx += len(write)

    @runs_in_hwd_thread
    def hid_read_frame(self):
        # INIT response
        read = bytearray(self.dbb_hid.read(self.usbReportSize))
        cid = ((read[0] * 256 + read[1]) * 256 + read[2]) * 256 + read[3]
        cmd = read[4]
        data_len = read[5] * 256 + read[6]
        data = read[7:]
        idx = len(read) - 7
        while idx < data_len:
            # CONT response
            read = bytearray(self.dbb_hid.read(self.usbReportSize))
            data += read[5:]
            idx += len(read) - 5
        return data

    @runs_in_hwd_thread
    def hid_send_plain(self, msg):
        reply = ""
        try:
            serial_number = self.dbb_hid.get_serial_number_string()
            if "v2.0." in serial_number or "v1." in serial_number:
                hidBufSize = 4096
                self.dbb_hid.write('\0' + msg + '\0' * (hidBufSize - len(msg)))
                r = bytearray()
                while len(r) < hidBufSize:
                    r += bytearray(self.dbb_hid.read(hidBufSize))
            else:
                self.hid_send_frame(msg)
                r = self.hid_read_frame()
            r = r.rstrip(b' \t\r\n\0')
            r = r.replace(b"\0", b'')
            r = to_string(r, 'utf8')
            reply = json.loads(r)
        except Exception as e:
            _logger.info(f'Exception caught {repr(e)}')
        return reply

    @runs_in_hwd_thread
    def hid_send_encrypt(self, msg):
        sha256_byte_len = 32
        reply = ""
        try:
            encryption_key, authentication_key = derive_keys(self.password)
            msg = EncodeAES_bytes(encryption_key, msg)
            hmac_digest = hmac_oneshot(authentication_key, msg, hashlib.sha256)
            authenticated_msg = base64.b64encode(msg + hmac_digest)
            reply = self.hid_send_plain(authenticated_msg)
            if 'ciphertext' in reply:
                b64_unencoded = bytes(base64.b64decode(''.join(reply["ciphertext"]), validate=True))
                reply_hmac = b64_unencoded[-sha256_byte_len:]
                hmac_calculated = hmac_oneshot(authentication_key, b64_unencoded[:-sha256_byte_len], hashlib.sha256)
                if not hmac.compare_digest(reply_hmac, hmac_calculated):
                    raise Exception("Failed to validate HMAC")
                reply = DecodeAES_bytes(encryption_key, b64_unencoded[:-sha256_byte_len])
                reply = to_string(reply, 'utf8')
                reply = json.loads(reply)
            if 'error' in reply:
                self.password = None
        except Exception as e:
            _logger.info(f'Exception caught {repr(e)}')
        return reply



# ----------------------------------------------------------------------------------
#
#

class DigitalBitbox_KeyStore(Hardware_KeyStore):
    hw_type = 'digitalbitbox'
    device = 'DigitalBitbox'

    plugin: 'DigitalBitboxPlugin'

    def __init__(self, d):
        Hardware_KeyStore.__init__(self, d)
        self.maxInputs = 14 # maximum inputs per single sign command

    def give_error(self, message: str | BaseException):
        raise Exception(message)

    def decrypt_message(self, pubkey, message, password):
        raise RuntimeError(_('Encryption and decryption are currently not supported for {}').format(self.device))

    def sign_message(self, sequence, message, password, *, script_type=None):
        sig = None
        try:
            message = message.encode('utf8')
            inputPath = self.get_derivation_prefix() + "/%d/%d" % sequence
            inputPath = normalize_bip32_derivation(inputPath, hardened_char="'")
            msg_hash = sha256d(usermessage_magic(message))
            inputHash = to_hexstr(msg_hash)
            hasharray = []
            hasharray.append({'hash': inputHash, 'keypath': inputPath})
            hasharray = json.dumps(hasharray)

            msg = ('{"sign":{"meta":"sign message", "data":%s}}' % hasharray).encode('utf8')

            dbb_client = self.plugin.get_client(self)

            if not dbb_client.is_paired():
                raise Exception(_("Could not sign message."))

            reply = dbb_client.hid_send_encrypt(msg)
            self.handler.show_message(_("Signing message ...") + "\n\n" +
                                      _("To continue, touch the Digital Bitbox's blinking light for 3 seconds.") + "\n\n" +
                                      _("To cancel, briefly touch the blinking light or wait for the timeout."))
            reply = dbb_client.hid_send_encrypt(msg) # Send twice, first returns an echo for smart verification (not implemented)
            self.handler.finished()

            if 'error' in reply:
                raise Exception(reply['error']['message'])

            if 'sign' not in reply:
                raise Exception(_("Could not sign message."))

            if 'recid' in reply['sign'][0]:
                # firmware > v2.1.1
                sig_string = binascii.unhexlify(reply['sign'][0]['sig'])
                recid = int(reply['sign'][0]['recid'], 16)
                sig = ecc.construct_ecdsa_sig65(sig_string, recid, is_compressed=True)
                pubkey, compressed, txin_type_guess = ecc.ECPubkey.from_ecdsa_sig65(sig, msg_hash)
                addr = public_key_to_p2pkh(pubkey.get_public_key_bytes(compressed=compressed))
                if verify_usermessage_with_address(addr, sig, message) is False:
                    raise Exception(_("Could not sign message"))
            elif 'pubkey' in reply['sign'][0]:
                # firmware <= v2.1.1
                for recid in range(4):
                    sig_string = binascii.unhexlify(reply['sign'][0]['sig'])
                    sig = ecc.construct_ecdsa_sig65(sig_string, recid, is_compressed=True)
                    try:
                        addr = public_key_to_p2pkh(binascii.unhexlify(reply['sign'][0]['pubkey']))
                        if verify_usermessage_with_address(addr, sig, message):
                            break
                    except Exception:
                        continue
                else:
                    raise Exception(_("Could not sign message"))

        except BaseException as e:
            self.give_error(e)
        return sig

    def sign_transaction(self, tx, password):
        if tx.is_complete():
            return

        try:
            p2pkhTransaction = True
            inputhasharray = []
            hasharray = []
            pubkeyarray = []

            # Build hasharray from inputs
            for i, txin in enumerate(tx.inputs()):
                if txin.is_coinbase_input():
                    self.give_error("Coinbase not supported") # should never happen

                desc = txin.script_descriptor
                assert desc
                if desc.to_legacy_electrum_script_type() != 'p2pkh':
                    p2pkhTransaction = False

                my_pubkey, inputPath = self.find_my_pubkey_in_txinout(txin)
                if not inputPath:
                    self.give_error("No matching pubkey for sign_transaction")  # should never happen
                inputPath = convert_bip32_intpath_to_strpath(inputPath)
                inputHash = sha256d(tx.serialize_preimage(i))
                hasharray_i = {'hash': to_hexstr(inputHash), 'keypath': inputPath}
                hasharray.append(hasharray_i)
                inputhasharray.append(inputHash)

            # Build pubkeyarray from outputs
            for txout in tx.outputs():
                assert txout.address
                if txout.is_change:
                    changePubkey, changePath = self.find_my_pubkey_in_txinout(txout)
                    assert changePath
                    changePath = convert_bip32_intpath_to_strpath(changePath)
                    changePubkey = changePubkey.hex()
                    pubkeyarray_i = {'pubkey': changePubkey, 'keypath': changePath}
                    pubkeyarray.append(pubkeyarray_i)

            # Special serialization of the unsigned transaction for
            # the mobile verification app.
            # At the moment, verification only works for p2pkh transactions.
            if p2pkhTransaction:
                tx_copy = copy.deepcopy(tx)
                # monkey-patch method of tx_copy instance to change serialization
                def input_script(self, txin: PartialTxInput, *, estimate_size=False) -> bytes:
                    desc = txin.script_descriptor
                    if isinstance(desc, descriptor.PKHDescriptor):
                        return txin.get_scriptcode_for_sighash()
                    raise Exception(f"unsupported txin type. only p2pkh is supported. got: {desc.to_string()[:10]}")
                tx_copy.input_script = input_script.__get__(tx_copy, PartialTransaction)
                tx_dbb_serialized = tx_copy.serialize_to_network()
            else:
                # We only need this for the signing echo / verification.
                tx_dbb_serialized = None

            # Build sign command
            dbb_signatures = []
            steps = math.ceil(1.0 * len(hasharray) / self.maxInputs)
            for step in range(int(steps)):
                hashes = hasharray[step * self.maxInputs : (step + 1) * self.maxInputs]

                msg = {
                    "sign": {
                        "data": hashes,
                        "checkpub": pubkeyarray,
                    },
                }
                if tx_dbb_serialized is not None:
                    msg["sign"]["meta"] = to_hexstr(sha256d(tx_dbb_serialized))
                msg = json.dumps(msg).encode('ascii')
                dbb_client = self.plugin.get_client(self)

                if not dbb_client.is_paired():
                    raise Exception("Could not sign transaction.")

                reply = dbb_client.hid_send_encrypt(msg)
                if 'error' in reply:
                    raise Exception(reply['error']['message'])

                if 'echo' not in reply:
                    raise Exception("Could not sign transaction.")

                if self.plugin.is_mobile_paired() and tx_dbb_serialized is not None:
                    reply['tx'] = tx_dbb_serialized
                    self.plugin.comserver_post_notification(reply, handler=self.handler)

                if steps > 1:
                    self.handler.show_message(_("Signing large transaction. Please be patient ...") + "\n\n" +
                                              _("To continue, touch the Digital Bitbox's blinking light for 3 seconds.") + " " +
                                              _("(Touch {} of {})").format((step + 1), steps) + "\n\n" +
                                              _("To cancel, briefly touch the blinking light or wait for the timeout.") + "\n\n")
                else:
                    self.handler.show_message(_("Signing transaction...") + "\n\n" +
                                              _("To continue, touch the Digital Bitbox's blinking light for 3 seconds.") + "\n\n" +
                                              _("To cancel, briefly touch the blinking light or wait for the timeout."))

                # Send twice, first returns an echo for smart verification
                reply = dbb_client.hid_send_encrypt(msg)
                self.handler.finished()

                if 'error' in reply:
                    if reply["error"].get('code') in (600, 601):
                        # aborted via LED short touch or timeout
                        raise UserCancelled()
                    raise Exception(reply['error']['message'])

                if 'sign' not in reply:
                    raise Exception("Could not sign transaction.")

                dbb_signatures.extend(reply['sign'])

            # Fill signatures
            if len(dbb_signatures) != len(tx.inputs()):
                raise Exception("Incorrect number of transactions signed.") # Should never occur
            for i, txin in enumerate(tx.inputs()):
                for pubkey_bytes in txin.pubkeys:
                    if txin.is_complete():
                        break
                    signed = dbb_signatures[i]
                    if 'recid' in signed:
                        # firmware > v2.1.1
                        recid = int(signed['recid'], 16)
                        s = binascii.unhexlify(signed['sig'])
                        h = inputhasharray[i]
                        pk = ecc.ECPubkey.from_ecdsa_sig64(s, recid, h)
                        pk = pk.get_public_key_hex(compressed=True)
                    elif 'pubkey' in signed:
                        # firmware <= v2.1.1
                        pk = signed['pubkey']
                    if pk != pubkey_bytes.hex():
                        continue
                    sig_r = int(signed['sig'][:64], 16)
                    sig_s = int(signed['sig'][64:], 16)
                    sig = ecc.ecdsa_der_sig_from_r_and_s(sig_r, sig_s)
                    sig = sig + Sighash.to_sigbytes(Sighash.ALL)
                    tx.add_signature_to_txin(txin_idx=i, signing_pubkey=pubkey_bytes, sig=sig)
        except UserCancelled:
            raise
        except BaseException as e:
            self.give_error(e)
        else:
            _logger.info(f"Transaction is_complete {tx.is_complete()}")


class DigitalBitboxPlugin(HW_PluginBase):

    libraries_available = DIGIBOX
    keystore_class = DigitalBitbox_KeyStore
    DEVICE_IDS = [
                   (0x03eb, 0x2402) # Digital Bitbox
                 ]
    SUPPORTED_XTYPES = ('standard', 'p2wpkh-p2sh', 'p2wpkh', 'p2wsh-p2sh', 'p2wsh')

    def __init__(self, parent, config, name):
        HW_PluginBase.__init__(self, parent, config, name)
        if self.libraries_available:
            self.device_manager().register_devices(self.DEVICE_IDS, plugin=self)

        self.digitalbitbox_config = self.config.get('digitalbitbox', {})

    @runs_in_hwd_thread
    def get_dbb_device(self, device):
        dev = hid.device()
        dev.open_path(device.path)
        return dev

    def create_client(self, device, handler):
        if device.interface_number == 0 or device.usage_page == 0xffff:
            client = self.get_dbb_device(device)
            if client is not None:
                client = DigitalBitbox_Client(self, client)
            return client
        else:
            return None

    def is_mobile_paired(self):
        return ENCRYPTION_PRIVKEY_KEY in self.digitalbitbox_config

    def comserver_post_notification(self, payload, *, handler: 'HardwareHandlerBase'):
        assert self.is_mobile_paired(), "unexpected mobile pairing error"
        url = 'https://digitalbitbox.com/smartverification/index.php'
        key_s = base64.b64decode(self.digitalbitbox_config[ENCRYPTION_PRIVKEY_KEY], validate=True)
        ciphertext = EncodeAES_bytes(key_s, json.dumps(payload).encode('ascii'))
        args = 'c=data&s=0&dt=0&uuid=%s&pl=%s' % (
            self.digitalbitbox_config[CHANNEL_ID_KEY],
            base64.b64encode(ciphertext).decode('ascii'),
        )
        try:
            text = Network.send_http_on_proxy('post', url, body=args.encode('ascii'), headers={'content-type': 'application/x-www-form-urlencoded'})
            _logger.info(f'digitalbitbox reply from server {text}')
        except Exception as e:
            _logger.exception("")
            handler.show_error(repr(e))  # repr because str(Exception()) == ''

    def get_client(self, keystore, force_pair=True, *,
                   devices=None, allow_user_interaction=True):
        client = super().get_client(keystore, force_pair,
                                    devices=devices,
                                    allow_user_interaction=allow_user_interaction)
        if client is not None:
            client.check_device_dialog()
        return client

    def show_address(self, wallet, address, keystore=None):
        if keystore is None:
            keystore = wallet.get_keystore()
        if not self.show_address_helper(wallet, address, keystore):
            return
        if type(wallet) is not Standard_Wallet:
            keystore.handler.show_error(_('This function is only available for standard wallets when using {}.').format(self.device))
            return
        if not self.is_mobile_paired():
            keystore.handler.show_error(_('This function is only available after pairing your {} with a mobile device.').format(self.device))
            return
        if wallet.get_txin_type(address) != 'p2pkh':
            keystore.handler.show_error(_('This function is only available for p2pkh keystores when using {}.').format(self.device))
            return
        change, index = wallet.get_address_index(address)
        keypath = '%s/%d/%d' % (keystore.get_derivation_prefix(), change, index)
        xpub = self.get_client(keystore)._get_xpub(keypath)
        verify_request_payload = {
            "type": 'p2pkh',
            "echo": xpub['echo'],
        }
        self.comserver_post_notification(verify_request_payload, handler=keystore.handler)

    def wizard_entry_for_device(self, device_info: 'DeviceInfo', *, new_wallet=True) -> str:
        if new_wallet:
            return 'dbitbox_start'
        else:
            return 'dbitbox_unlock'

    # insert digitalbitbox pages in new wallet wizard
    def extend_wizard(self, wizard: 'NewWalletWizard'):
        views = {
            'dbitbox_start': {
                'next': 'dbitbox_xpub',
            },
            'dbitbox_xpub': {
                'next': lambda d: wizard.wallet_password_view(d) if wizard.last_cosigner(d) else 'multisig_cosigner_keystore',
                'accept': wizard.maybe_master_pubkey,
                'last': lambda d: wizard.is_single_password() and wizard.last_cosigner(d)
            },
            'dbitbox_unlock': {
                'last': True
            },
        }
        wizard.navmap_merge(views)



import copy
import threading
from abc import abstractmethod
from typing import TYPE_CHECKING, Optional

from PyQt6.QtCore import Qt, QTimer, pyqtSignal, pyqtSlot, QSize, QMetaObject
from PyQt6.QtGui import QPixmap
from PyQt6.QtWidgets import (QDialog, QPushButton, QWidget, QLabel, QVBoxLayout, QScrollArea,
                             QHBoxLayout, QLayout)

from electrum.i18n import _
from electrum.logging import get_logger
from electrum.gui.qt.util import Buttons, icon_path, MessageBoxMixin, WWLabel, ResizableStackedWidget, AbstractQWidget

if TYPE_CHECKING:
    from electrum.simple_config import SimpleConfig
    from electrum.gui.qt import QElectrumApplication
    from electrum.wizard import WizardViewState


class QEAbstractWizard(QDialog, MessageBoxMixin):
    """ Concrete subclasses of QEAbstractWizard must also inherit from a concrete AbstractWizard subclass.
        QEAbstractWizard forms the base for all QtWidgets GUI based wizards, while AbstractWizard defines
        the base for non-gui wizard flow navigation functionality.
    """
    _logger = get_logger(__name__)

    requestNext = pyqtSignal()
    requestPrev = pyqtSignal()

    def __init__(self, config: 'SimpleConfig', app: 'QElectrumApplication', *, start_viewstate: 'WizardViewState' = None):
        QDialog.__init__(self, None)
        self.app = app
        self.config = config

        # compat
        self.gui_thread = threading.current_thread()

        self.setMinimumSize(600, 400)

        self.title = QLabel()
        self.window_title = ''
        self.finish_label = _('Finish')

        self.main_widget = ResizableStackedWidget(self)

        self.back_button = QPushButton(_("Back"), self)
        self.back_button.clicked.connect(self.on_back_button_clicked)
        self.back_button.setEnabled(False)
        self.back_button.setDefault(False)
        self.back_button.setAutoDefault(False)
        self.next_button = QPushButton(_("Next"), self)
        self.next_button.clicked.connect(self.on_next_button_clicked)
        self.next_button.setEnabled(False)
        self.next_button.setDefault(True)
        self.next_button.setAutoDefault(True)
        self.requestPrev.connect(self.on_back_button_clicked)
        self.requestNext.connect(self.on_next_button_clicked)
        self.logo = QLabel()

        please_wait_layout = QVBoxLayout()
        please_wait_layout.addStretch(1)
        self.please_wait_l = QLabel(_("Please wait..."))
        self.please_wait_l.setAlignment(Qt.AlignmentFlag.AlignCenter)
        please_wait_layout.addWidget(self.please_wait_l)
        please_wait_layout.addStretch(1)
        self.please_wait = QWidget()
        self.please_wait.setVisible(False)
        self.please_wait.setLayout(please_wait_layout)

        error_layout = QVBoxLayout()
        error_layout.addStretch(1)
        error_icon = QLabel()
        error_icon.setPixmap(QPixmap(icon_path('warning.png')).scaledToWidth(48, mode=Qt.TransformationMode.SmoothTransformation))
        error_icon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        error_layout.addWidget(error_icon)
        self.error_msg = WWLabel()
        self.error_msg.setAlignment(Qt.AlignmentFlag.AlignCenter)
        error_layout.addWidget(self.error_msg)
        error_layout.addStretch(1)
        self.error = QWidget()
        self.error.setVisible(False)
        self.error.setLayout(error_layout)

        outer_vbox = QVBoxLayout(self)
        inner_vbox = QVBoxLayout()
        inner_vbox.addWidget(self.title)
        inner_vbox.addWidget(self.main_widget)
        inner_vbox.addWidget(self.please_wait)
        inner_vbox.addWidget(self.error)

        scroll_widget = QWidget()
        scroll_widget.setLayout(inner_vbox)
        scroll = QScrollArea()
        scroll.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        scroll.setWidget(scroll_widget)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        scroll.setWidgetResizable(True)
        icon_vbox = QVBoxLayout()
        icon_vbox.addWidget(self.logo)
        icon_vbox.addStretch(1)
        hbox = QHBoxLayout()
        hbox.addLayout(icon_vbox)
        hbox.addSpacing(5)
        hbox.addWidget(scroll)
        hbox.setStretchFactor(scroll, 1)
        outer_vbox.addLayout(hbox)
        outer_vbox.addLayout(Buttons(self.back_button, self.next_button))

        self.setTabOrder(self.back_button, self.next_button)

        self.icon_filename = None
        self.set_icon('electrum.png')

        self.start_viewstate = start_viewstate

        self.show()
        self.raise_()

        QMetaObject.invokeMethod(self, 'strt', Qt.ConnectionType.QueuedConnection)  # call strt after subclass constructor(s)

    def sizeHint(self) -> QSize:
        return QSize(600, 400)

    @pyqtSlot()
    def strt(self):
        viewstate = self.start_wizard(start_viewstate=self.start_viewstate)
        self.load_next_component(viewstate.view, viewstate.wizard_data, viewstate.params)
        self.set_default_focus()

        # TODO: re-test if needed on macOS
        self.refresh_gui()  # Need for QT on MacOSX.  Lame.

    def refresh_gui(self):
        # For some reason, to refresh the GUI this needs to be called twice
        self.app.processEvents()
        self.app.processEvents()

    def load_next_component(self, view, wdata=None, params=None):
        if wdata is None:
            wdata = {}
        if params is None:
            params = {}

        comp = self.view_to_component(view)
        try:
            self._logger.debug(f'load_next_component: {comp!r}')
            page = comp(self.main_widget, self)
        except Exception as e:
            self._logger.error(f'not a class: {comp!r}')
            raise e
        page.wizard_data = copy.deepcopy(wdata)
        page.params = params
        page.on_ready()  # call before component emits any signals

        page.updated.connect(self.on_page_updated)

        # add to stack and update wizard
        page.apply()
        self.main_widget.setCurrentIndex(self.main_widget.addWidget(page))
        self.update()

    @pyqtSlot(object)
    def on_page_updated(self, page):
        page.apply()
        if page == self.main_widget.currentWidget():
            self.update()

    def set_icon(self, filename):
        prior_filename, self.icon_filename = self.icon_filename, filename
        self.logo.setPixmap(QPixmap(icon_path(filename))
                            .scaledToWidth(60, mode=Qt.TransformationMode.SmoothTransformation))
        return prior_filename

    def set_default_focus(self):
        page = self.main_widget.currentWidget()
        control = page.initialFocus()
        if control and control.isVisible() and control.isEnabled():
            control.setFocus()
        else:
            self.next_button.setFocus()

    def can_go_back(self) -> bool:
        return len(self._stack) > 0

    def update(self):
        page = self.main_widget.currentWidget()
        self.setWindowTitle(page.wizard_title if page.wizard_title else self.window_title)
        self.title.setText(f'<b>{page.title}</b>' if page.title else '')
        self.back_button.setText(_('Back') if self.can_go_back() else _('Cancel'))
        self.back_button.setEnabled(not page.busy)
        self.next_button.setText(_('Next') if not self.is_last(page.wizard_data) else self.finish_label)
        self.next_button.setEnabled(not page.busy and page.valid)
        self.main_widget.setVisible(not page.busy and not bool(page.error))
        self.please_wait.setVisible(page.busy)
        self.please_wait_l.setText(page.busy_msg if page.busy_msg else _("Please wait..."))
        self.error_msg.setText(str(page.error))
        self.error.setVisible(not page.busy and bool(page.error))
        icon = page.params.get('icon', icon_path('electrum.png'))
        if icon:
            if icon != self.icon_filename:
                self.set_icon(icon)
            self.logo.setVisible(True)
        else:
            self.logo.setVisible(False)

    def on_back_button_clicked(self):
        if self.can_go_back():
            self.prev()
            widget = self.main_widget.currentWidget()
            self.main_widget.removeWidget(widget)
            widget.deleteLater()
            self.update()
        else:
            self.close()

    def on_next_button_clicked(self):
        page = self.main_widget.currentWidget()
        page.apply()
        wd = page.wizard_data.copy()
        if self.is_last(wd):
            self.submit(wd)
            if self.is_finalized(wd):
                self.accept()
            else:
                self.prev()  # rollback the submit above
        else:
            view = self.submit(wd)
            try:
                self.load_next_component(view.view, view.wizard_data, view.params)
                self.set_default_focus()
            except Exception as e:
                self.prev()  # rollback the submit above
                raise e

    def start_wizard(self, *, start_viewstate: Optional['WizardViewState'] = None) -> 'WizardViewState':
        self.start(start_viewstate=start_viewstate)
        return self._current

    def view_to_component(self, view) -> QWidget:
        return self.navmap[view]['gui']

    def submit(self, wizard_data) -> 'WizardViewState':
        wdata = wizard_data.copy()
        view = self.resolve_next(self._current.view, wdata)
        return view

    def prev(self) -> dict:
        viewstate = self.resolve_prev()
        return viewstate.wizard_data

    def is_last(self, wizard_data: dict) -> bool:
        wdata = wizard_data.copy()
        return self.is_last_view(self._current.view, wdata)

    def is_finalized(self, wizard_data: dict) -> bool:
        ''' Final check before closing the wizard. '''
        return True


class WizardComponent(AbstractQWidget):
    updated = pyqtSignal(object)

    def __init__(self, parent: QWidget, wizard: QEAbstractWizard, *, title: str = None, layout: QLayout = None):
        super().__init__(parent)
        self.setLayout(layout if layout else QVBoxLayout(self))
        self.wizard_data = {}
        self.title = title if title is not None else 'No title'
        self.wizard_title = None
        self.busy_msg = ''
        self.wizard = wizard
        self._error = ''
        self._valid = False
        self._busy = False

    @property
    def valid(self):
        return self._valid

    @valid.setter
    def valid(self, is_valid):
        if self._valid != is_valid:
            self._valid = is_valid
            self.on_updated()

    @property
    def busy(self):
        return self._busy

    @busy.setter
    def busy(self, is_busy):
        if self._busy != is_busy:
            self._busy = is_busy
            self.on_updated()

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, error):
        if self._error != error:
            self._error = error
            self.on_updated()

    @abstractmethod
    def apply(self):
        # called to apply UI component values to wizard_data
        pass

    def on_ready(self):
        # called when wizard_data is available
        pass

    @pyqtSlot()
    def on_updated(self, *args):
        try:
            self.updated.emit(self)
        except RuntimeError:
            pass

    def initialFocus(self) -> Optional[QWidget]:
        """Override to specify a control that should receive initial focus"""
        return None



#!/usr/bin/env python3

# A simple script that connects to a server and displays block headers

import time
import asyncio

from electrum.network import Network
from electrum.util import print_msg, json_encode, create_and_start_event_loop, log_exceptions
from electrum.simple_config import SimpleConfig

config = SimpleConfig()

# start network
loop, stopping_fut, loop_thread = create_and_start_event_loop()
network = Network(config)
network.start()

# wait until connected
while not network.is_connected():
    time.sleep(1)
    print_msg("waiting for network to get connected...")

header_queue = asyncio.Queue()

@log_exceptions
async def f():
    try:
        await network.interface.session.subscribe('blockchain.headers.subscribe', [], header_queue)
        # 3. wait for results
        while network.is_connected():
            header = await header_queue.get()
            print_msg(json_encode(header))
    finally:
        stopping_fut.set_result(1)

# 2. send the subscription
asyncio.run_coroutine_threadsafe(f(), loop)
while loop_thread.is_alive():
    loop_thread.join(1)



# -*- coding: utf-8 -*-
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2018 The Electrum developers
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import asyncio
import threading
from typing import TYPE_CHECKING, Dict, Set

import aiorpcx
import electrum_ecc as ecc
from electrum_ecc import ECPubkey

from . import bitcoin
from . import constants
from .util import bfh, NetworkJobOnDefaultServer
from .lnutil import funding_output_script_from_keys, ShortChannelID
from .verifier import verify_tx_is_in_block, MerkleVerificationFailure
from .transaction import Transaction
from .interface import GracefulDisconnect
from .crypto import sha256d
from .lnmsg import decode_msg, encode_msg

if TYPE_CHECKING:
    from .network import Network
    from .lnrouter import ChannelDB


class LNChannelVerifier(NetworkJobOnDefaultServer):
    """ Verify channel announcements for the Channel DB """

    # FIXME the initial routing sync is bandwidth-heavy, and the electrum server
    # will start throttling us, making it even slower. one option would be to
    # spread it over multiple servers.

    def __init__(self, network: 'Network', channel_db: 'ChannelDB'):
        self.channel_db = channel_db
        self.lock = threading.Lock()
        self.unverified_channel_info = {}  # type: Dict[ShortChannelID, dict]  # scid -> msg_dict
        # channel announcements that seem to be invalid:
        self.blacklist = set()  # type: Set[ShortChannelID]
        NetworkJobOnDefaultServer.__init__(self, network)

    def _reset(self):
        super()._reset()
        self.started_verifying_channel = set()  # type: Set[ShortChannelID]

    # TODO make async; and rm self.lock completely
    def add_new_channel_info(self, short_channel_id: ShortChannelID, msg: dict) -> bool:
        if short_channel_id in self.unverified_channel_info:
            return False
        if short_channel_id in self.blacklist:
            return False
        with self.lock:
            self.unverified_channel_info[short_channel_id] = msg
            return True

    async def _run_tasks(self, *, taskgroup):
        await super()._run_tasks(taskgroup=taskgroup)
        async with taskgroup as group:
            await group.spawn(self.main)

    async def main(self):
        while True:
            await self._verify_some_channels()
            await asyncio.sleep(0.1)

    async def _verify_some_channels(self):
        blockchain = self.network.blockchain()
        local_height = blockchain.height()

        with self.lock:
            unverified_channel_info = list(self.unverified_channel_info)

        for short_channel_id in unverified_channel_info:
            if short_channel_id in self.started_verifying_channel:
                continue
            block_height = short_channel_id.block_height
            # only resolve short_channel_id if headers are available.
            if block_height <= 0 or block_height > local_height:
                continue
            header = blockchain.read_header(block_height)
            if header is None:
                if block_height <= constants.net.max_checkpoint():
                    await self.taskgroup.spawn(self.interface.request_chunk_below_max_checkpoint(height=block_height))
                continue
            self.started_verifying_channel.add(short_channel_id)
            await self.taskgroup.spawn(self.verify_channel(block_height, short_channel_id))
            #self.logger.info(f'requested short_channel_id {short_channel_id.hex()}')

    async def verify_channel(self, block_height: int, short_channel_id: ShortChannelID):
        # we are verifying channel announcements as they are from untrusted ln peers.
        # we use electrum servers to do this. however we don't trust electrum servers either...
        try:
            async with self._network_request_semaphore:
                result = await self.network.get_txid_from_txpos(
                    block_height, short_channel_id.txpos, True)
        except aiorpcx.jsonrpc.RPCError:
            # the electrum server is complaining about the txpos for given block.
            # it is not clear what to do now, but let's believe the server.
            self._blacklist_short_channel_id(short_channel_id)
            return
        tx_hash = result['tx_hash']
        merkle_branch = result['merkle']
        # we need to wait if header sync/reorg is still ongoing, hence lock:
        async with self.network.bhi_lock:
            header = self.network.blockchain().read_header(block_height)
        try:
            verify_tx_is_in_block(tx_hash, merkle_branch, short_channel_id.txpos, header, block_height)
        except MerkleVerificationFailure as e:
            # the electrum server sent an incorrect proof. blame is on server, not the ln peer
            raise GracefulDisconnect(e) from e
        try:
            async with self._network_request_semaphore:
                raw_tx = await self.network.get_transaction(tx_hash)
        except aiorpcx.jsonrpc.RPCError as e:
            # the electrum server can't find the tx; but it was the
            # one who told us about the txid!! blame is on server
            raise GracefulDisconnect(e) from e
        tx = Transaction(raw_tx)
        try:
            tx.deserialize()
        except Exception:
            # either bug in client, or electrum server is evil.
            # if we connect to a diff server at some point, let's try again.
            self.logger.warning(f"cannot deserialize transaction, skipping {tx_hash}")
            return
        if tx_hash != tx.txid():
            # either bug in client, or electrum server is evil.
            # if we connect to a diff server at some point, let's try again.
            self.logger.info(f"received tx does not match expected txid ({tx_hash} != {tx.txid()})")
            return
        # check funding output
        chan_ann_msg = self.unverified_channel_info[short_channel_id]
        redeem_script = funding_output_script_from_keys(chan_ann_msg['bitcoin_key_1'], chan_ann_msg['bitcoin_key_2'])
        expected_address = bitcoin.redeem_script_to_address('p2wsh', redeem_script)
        try:
            actual_output = tx.outputs()[short_channel_id.output_index]
        except IndexError:
            self._blacklist_short_channel_id(short_channel_id)
            return
        if expected_address != actual_output.address:
            # FIXME what now? best would be to ban the originating ln peer.
            self.logger.info(f"funding output script mismatch for {short_channel_id}")
            self._remove_channel_from_unverified_db(short_channel_id)
            return
        # put channel into channel DB
        self.channel_db.add_verified_channel_info(chan_ann_msg, capacity_sat=actual_output.value)
        self._remove_channel_from_unverified_db(short_channel_id)

    def _remove_channel_from_unverified_db(self, short_channel_id: ShortChannelID):
        with self.lock:
            self.unverified_channel_info.pop(short_channel_id, None)
        self.started_verifying_channel.discard(short_channel_id)

    def _blacklist_short_channel_id(self, short_channel_id: ShortChannelID) -> None:
        self.blacklist.add(short_channel_id)
        with self.lock:
            self.unverified_channel_info.pop(short_channel_id, None)


def verify_sig_for_channel_update(chan_upd: dict, node_id: bytes) -> bool:
    msg_bytes = chan_upd['raw']
    pre_hash = msg_bytes[2+64:]
    h = sha256d(pre_hash)
    sig = chan_upd['signature']
    if not ECPubkey(node_id).ecdsa_verify(sig, h):
        return False
    return True



import threading
import socket
import base64
from typing import TYPE_CHECKING

from PyQt6.QtCore import pyqtSignal, pyqtProperty, pyqtSlot

from electrum.i18n import _
from electrum.bip32 import BIP32Node
from electrum import bitcoin

from .trustedcoin import (server, ErrorConnectingServer, MOBILE_DISCLAIMER, TrustedCoinException)
from electrum.gui.common_qt.plugins import PluginQObject

if TYPE_CHECKING:
    from electrum.wizard import NewWalletWizard


class TrustedcoinPluginQObject(PluginQObject):
    canSignWithoutServerChanged = pyqtSignal()
    termsAndConditionsRetrieved = pyqtSignal([str], arguments=['message'])
    termsAndConditionsError = pyqtSignal([str], arguments=['message'])
    otpError = pyqtSignal([str], arguments=['message'])
    otpSuccess = pyqtSignal()
    disclaimerChanged = pyqtSignal()
    keystoreChanged = pyqtSignal()
    otpSecretChanged = pyqtSignal()
    shortIdChanged = pyqtSignal()
    billingModelChanged = pyqtSignal()

    remoteKeyStateChanged = pyqtSignal()
    remoteKeyError = pyqtSignal([str], arguments=['message'])

    requestOtp = pyqtSignal()

    def __init__(self, plugin, wizard: 'NewWalletWizard', parent):
        super().__init__(plugin, parent)
        self.wizard = wizard
        self._canSignWithoutServer = False
        self._otpSecret = ''
        self._shortId = ''
        self._billingModel = []
        self._remoteKeyState = ''
        self._verifyingOtp = False

    @pyqtProperty(str, notify=disclaimerChanged)
    def disclaimer(self):
        return '\n\n'.join(MOBILE_DISCLAIMER)

    @pyqtProperty(bool, notify=canSignWithoutServerChanged)
    def canSignWithoutServer(self):
        return self._canSignWithoutServer

    @pyqtProperty('QVariantMap', notify=keystoreChanged)
    def keystore(self):
        return self._keystore

    @pyqtProperty(str, notify=otpSecretChanged)
    def otpSecret(self):
        return self._otpSecret

    @pyqtProperty(str, notify=shortIdChanged)
    def shortId(self):
        return self._shortId

    @pyqtSlot(str)
    def otpSubmit(self, otp):
        self._plugin.on_otp(otp)

    @pyqtProperty(str, notify=remoteKeyStateChanged)
    def remoteKeyState(self):
        return self._remoteKeyState

    @remoteKeyState.setter
    def remoteKeyState(self, new_state):
        if self._remoteKeyState != new_state:
            self._remoteKeyState = new_state
            self.remoteKeyStateChanged.emit()

    @pyqtProperty('QVariantList', notify=billingModelChanged)
    def billingModel(self):
        return self._billingModel

    def updateBillingInfo(self, wallet):
        billing_model = []

        price_per_tx = wallet.price_per_tx
        for k, v in sorted(price_per_tx.items()):
            if k == 1:
                continue
            item = {
                'text': 'Pay every %d transactions' % k,
                'value': k,
                'sats_per_tx': v / k
            }
            billing_model.append(item)

        self._billingModel = billing_model
        self.billingModelChanged.emit()

    @pyqtSlot()
    def fetchTermsAndConditions(self):
        def fetch_task():
            try:
                self.plugin.logger.debug('TOS')
                tos = server.get_terms_of_service()
            except ErrorConnectingServer as e:
                self.termsAndConditionsError.emit(_('Error connecting to server'))
            except Exception as e:
                self.termsAndConditionsError.emit('%s: %s' % (_('Error'), repr(e)))
            else:
                self.termsAndConditionsRetrieved.emit(tos)
            finally:
                self._busy = False
                self.busyChanged.emit()

        self._busy = True
        self.busyChanged.emit()
        t = threading.Thread(target=fetch_task)
        t.daemon = True
        t.start()

    @pyqtSlot()
    def createKeystore(self):
        email = 'dummy@electrum.org'

        self.remoteKeyState = ''
        self._otpSecret = ''
        self.otpSecretChanged.emit()

        wizard_data = self.wizard.get_wizard_data()

        xprv1, xpub1, xprv2, xpub2, xpub3, short_id = self.plugin.create_keys(wizard_data)

        def create_remote_key_task():
            try:
                self.plugin.logger.debug('create remote key')
                r = server.create(xpub1, xpub2, email)

                otp_secret = r['otp_secret']
                _xpub3 = r['xpubkey_cosigner']
                _id = r['id']
            except (socket.error, ErrorConnectingServer) as e:
                self.remoteKeyState = 'error'
                self.remoteKeyError.emit(f'Network error: {str(e)}')
            except TrustedCoinException as e:
                if e.status_code == 409:
                    self.remoteKeyState = 'wallet_known'
                    self._shortId = short_id
                    self.shortIdChanged.emit()
                else:
                    self.remoteKeyState = 'error'
                    self.logger.warning(str(e))
                    self.remoteKeyError.emit(f'Service error: {str(e)}')
            except (KeyError, TypeError) as e:  # catch any assumptions
                self.remoteKeyState = 'error'
                self.remoteKeyError.emit(f'Error: {str(e)}')
                self.logger.error(str(e))
            else:
                if short_id != _id:
                    self.remoteKeyState = 'error'
                    self.logger.error("unexpected trustedcoin short_id: expected {}, received {}".format(short_id, _id))
                    self.remoteKeyError.emit('Unexpected short_id')
                    return
                if xpub3 != _xpub3:
                    self.remoteKeyState = 'error'
                    self.logger.error("unexpected trustedcoin xpub3: expected {}, received {}".format(xpub3, _xpub3))
                    self.remoteKeyError.emit('Unexpected trustedcoin xpub3')
                    return
                self.remoteKeyState = 'new'
                self._otpSecret = otp_secret
                self.otpSecretChanged.emit()
                self._shortId = short_id
                self.shortIdChanged.emit()
            finally:
                self._busy = False
                self.busyChanged.emit()

        self._busy = True
        self.busyChanged.emit()

        t = threading.Thread(target=create_remote_key_task)
        t.daemon = True
        t.start()

    @pyqtSlot()
    def resetOtpSecret(self):
        self.remoteKeyState = ''

        wizard_data = self.wizard.get_wizard_data()

        xprv1, xpub1, xprv2, xpub2, xpub3, short_id = self.plugin.create_keys(wizard_data)

        def reset_otp_task():
            try:
                self.plugin.logger.debug('reset_otp')
                r = server.get_challenge(short_id)
                challenge = r.get('challenge')
                message = 'TRUSTEDCOIN CHALLENGE: ' + challenge

                def f(xprv):
                    rootnode = BIP32Node.from_xkey(xprv)
                    key = rootnode.subkey_at_private_derivation((0, 0)).eckey
                    sig = bitcoin.ecdsa_sign_usermessage(key, message, is_compressed=True)
                    return base64.b64encode(sig).decode()

                signatures = [f(x) for x in [xprv1, xprv2]]
                r = server.reset_auth(short_id, challenge, signatures)
                otp_secret = r.get('otp_secret')
            except (socket.error, ErrorConnectingServer) as e:
                self.remoteKeyState = 'error'
                self.remoteKeyError.emit(f'Network error: {str(e)}')
            except Exception as e:
                self.remoteKeyState = 'error'
                self.remoteKeyError.emit(f'Error: {str(e)}')
            else:
                self.remoteKeyState = 'reset'
                self._otpSecret = otp_secret
                self.otpSecretChanged.emit()
            finally:
                self._busy = False
                self.busyChanged.emit()

        self._busy = True
        self.busyChanged.emit()

        t = threading.Thread(target=reset_otp_task, daemon=True)
        t.start()

    @pyqtSlot(str, int)
    def checkOtp(self, short_id, otp):
        assert type(otp) is int  # make sure this doesn't fail subtly

        def check_otp_task():
            try:
                self.plugin.logger.debug(f'check OTP, shortId={short_id}, otp={otp}')
                server.auth(short_id, otp)
            except TrustedCoinException as e:
                if e.status_code == 400:  # invalid OTP
                    self.plugin.logger.debug('Invalid one-time password.')
                    self.otpError.emit(_('Invalid one-time password.'))
                else:
                    self.plugin.logger.error(str(e))
                    self.otpError.emit(f'Service error: {str(e)}')
            except Exception as e:
                self.plugin.logger.error(str(e))
                self.otpError.emit(f'Error: {str(e)}')
            else:
                self.plugin.logger.debug('OTP verify success')
                self.otpSuccess.emit()
            finally:
                self._busy = False
                self.busyChanged.emit()
                self._verifyingOtp = False

        self._verifyingOtp = True
        self._busy = True
        self.busyChanged.emit()
        t = threading.Thread(target=check_otp_task, daemon=True)
        t.start()



# Copyright (C) 2020 The Electrum developers
# Distributed under the MIT software license, see the accompanying
# file LICENCE or http://www.opensource.org/licenses/mit-license.php
"""
lnrater.py contains Lightning Network node rating functionality.
"""

import asyncio
from collections import defaultdict
from pprint import pformat
from random import choices
from statistics import mean, median, stdev
from typing import TYPE_CHECKING, Dict, NamedTuple, Tuple, List, Optional
import sys
import time

from .logging import Logger
from .util import profiler, get_running_loop
from .lnrouter import fee_for_edge_msat
from .lnutil import LnFeatures, ln_compare_features, IncompatibleLightningFeatures
from .network import Network

if TYPE_CHECKING:
    from .channel_db import Policy, NodeInfo
    from .lnchannel import ShortChannelID
    from .lnworker import LNWallet


MONTH_IN_BLOCKS = 6 * 24 * 30
# the scores are only updated after this time interval
RATER_UPDATE_TIME_SEC = 10 * 60
# amount used for calculating an effective relative fee
FEE_AMOUNT_MSAT = 100_000_000

# define some numbers for minimal requirements of good nodes
# exclude nodes with less number of channels
EXCLUDE_NUM_CHANNELS = 15
# exclude nodes with less mean capacity
EXCLUDE_MEAN_CAPACITY_MSAT = 1_000_000_000
# exclude nodes which are young
EXCLUDE_NODE_AGE = 2 * MONTH_IN_BLOCKS
# exclude nodes which have young mean channel age
EXCLUDE_MEAN_CHANNEL_AGE = EXCLUDE_NODE_AGE
# exclude nodes which charge a high fee
EXCLUDE_EFFECTIVE_FEE_RATE = 0.001500
# exclude nodes whose last channel open was a long time ago
EXCLUDE_BLOCKS_LAST_CHANNEL = 3 * MONTH_IN_BLOCKS


class NodeStats(NamedTuple):
    number_channels: int
    # capacity related
    total_capacity_msat: int
    median_capacity_msat: float
    mean_capacity_msat: float
    # block height related
    node_age_block_height: int
    mean_channel_age_block_height: float
    blocks_since_last_channel: int
    # fees
    mean_fee_rate: float


def weighted_sum(numbers: List[float], weights: List[float]) -> float:
    running_sum = 0.0
    for n, w in zip(numbers, weights):
        running_sum += n * w
    return running_sum/sum(weights)


class LNRater(Logger):
    def __init__(self, lnworker: 'LNWallet', network: 'Network'):
        """LNRater can be used to suggest nodes to open up channels with.

        The graph is analyzed and some heuristics are applied to sort out nodes
        that are deemed to be bad routers or unmaintained.
        """
        Logger.__init__(self)
        self.lnworker = lnworker
        self.network = network

        self._node_stats: Dict[bytes, NodeStats] = {}  # node_id -> NodeStats
        self._node_ratings: Dict[bytes, float] = {}  # node_id -> float
        self._policies_by_nodes: Dict[bytes, List[Tuple[ShortChannelID, Policy]]] = defaultdict(list)  # node_id -> (short_channel_id, policy)
        self._last_analyzed = 0  # timestamp
        self._last_progress_percent = 0

    def maybe_analyze_graph(self):
        Network.run_from_another_thread(self._maybe_analyze_graph())

    def analyze_graph(self):
        """Forces a graph analysis, e.g., due to external triggers like
        the graph info reaching 50%."""
        Network.run_from_another_thread(self._analyze_graph())

    async def _maybe_analyze_graph(self):
        """Analyzes the graph when in early sync stage (>30%) or when caching
        time expires."""
        # gather information about graph sync status
        current_channels, total, progress_percent = self.network.lngossip.get_sync_progress_estimate()

        # gossip sync progress state could be None when not started, but channel
        # db already knows something about the graph, which is why we allow to
        # evaluate the graph early
        if progress_percent is not None or self.network.channel_db.num_nodes > 500:
            progress_percent = progress_percent or 0  # convert None to 0
            now = time.time()
            # graph should have changed significantly during the sync progress
            # or last analysis was a long time ago
            if (30 <= progress_percent and progress_percent - self._last_progress_percent >= 10 or
                    self._last_analyzed + RATER_UPDATE_TIME_SEC < now):
                await self._analyze_graph()
                self._last_progress_percent = progress_percent
                self._last_analyzed = now

    async def _analyze_graph(self):
        await self.network.channel_db.data_loaded.wait()
        self._collect_policies_by_node()
        loop = get_running_loop()
        # the analysis is run in an executor because it's costly
        await loop.run_in_executor(None, self._collect_purged_stats)
        self._rate_nodes()
        now = time.time()
        self._last_analyzed = now

    def _collect_policies_by_node(self):
        policies = self.network.channel_db.get_node_policies()
        for pv, p in policies.items():
            # append tuples of ShortChannelID and Policy
            self._policies_by_nodes[pv[0]].append((pv[1], p))

    @profiler
    def _collect_purged_stats(self):
        """Traverses through the graph and sorts out nodes."""
        current_height = self.network.get_local_height()
        node_infos = self.network.channel_db.get_node_infos()

        for n, channel_policies in self._policies_by_nodes.items():
            try:
                # use policies synonymously to channels
                num_channels = len(channel_policies)

                # save some time for nodes we are not interested in:
                if num_channels < EXCLUDE_NUM_CHANNELS:
                    continue

                # analyze block heights
                block_heights = [p[0].block_height for p in channel_policies]
                node_age_bh = current_height - min(block_heights)
                if node_age_bh < EXCLUDE_NODE_AGE:
                    continue
                mean_channel_age_bh = current_height - mean(block_heights)
                if mean_channel_age_bh < EXCLUDE_MEAN_CHANNEL_AGE:
                    continue
                blocks_since_last_channel = current_height - max(block_heights)
                if blocks_since_last_channel > EXCLUDE_BLOCKS_LAST_CHANNEL:
                    continue

                # analyze capacities
                capacities = [p[1].htlc_maximum_msat for p in channel_policies]
                if None in capacities:
                    continue
                total_capacity = sum(capacities)

                mean_capacity = total_capacity / num_channels if num_channels else 0
                if mean_capacity < EXCLUDE_MEAN_CAPACITY_MSAT:
                    continue
                median_capacity = median(capacities)

                # analyze fees
                effective_fee_rates = [fee_for_edge_msat(
                    FEE_AMOUNT_MSAT,
                    p[1].fee_base_msat,
                    p[1].fee_proportional_millionths) / FEE_AMOUNT_MSAT for p in channel_policies]
                mean_fees_rate = mean(effective_fee_rates)
                if mean_fees_rate > EXCLUDE_EFFECTIVE_FEE_RATE:
                    continue

                self._node_stats[n] = NodeStats(
                    number_channels=num_channels,
                    total_capacity_msat=total_capacity,
                    median_capacity_msat=median_capacity,
                    mean_capacity_msat=mean_capacity,
                    node_age_block_height=node_age_bh,
                    mean_channel_age_block_height=mean_channel_age_bh,
                    blocks_since_last_channel=blocks_since_last_channel,
                    mean_fee_rate=mean_fees_rate
                )

            except Exception as e:
                self.logger.exception("Could not use channel policies for "
                                      "calculating statistics.")
                self.logger.debug(pformat(channel_policies))
                continue

        self.logger.info(f"node statistics done, calculated statistics "
                         f"for {len(self._node_stats)} nodes")

    def _rate_nodes(self):
        """Rate nodes by collected statistics."""

        max_capacity = 0
        max_num_chan = 0
        min_fee_rate = float('inf')
        for stats in self._node_stats.values():
            max_capacity = max(max_capacity, stats.total_capacity_msat)
            max_num_chan = max(max_num_chan, stats.number_channels)
            min_fee_rate = min(min_fee_rate, stats.mean_fee_rate)

        for n, stats in self._node_stats.items():
            heuristics = []
            heuristics_weights = []

            # Construct an average score which leads to recommendation of nodes
            # with low fees, large capacity and reasonable number of channels.
            # This is somewhat akin to preferential attachment, but low fee
            # nodes are more favored. Here we make a compromise between user
            # comfort and decentralization, tending towards user comfort.

            # number of channels
            heuristics.append(stats.number_channels / max_num_chan)
            heuristics_weights.append(0.2)
            # total capacity
            heuristics.append(stats.total_capacity_msat / max_capacity)
            heuristics_weights.append(0.8)
            # inverse fees
            fees = min(1E-6, min_fee_rate) / max(1E-10, stats.mean_fee_rate)
            heuristics.append(fees)
            heuristics_weights.append(1.0)

            self._node_ratings[n] = weighted_sum(heuristics, heuristics_weights)

    @profiler
    def suggest_node_channel_open(self) -> Optional[bytes]:
        node_stats = self._node_stats.copy()
        node_ratings = self._node_ratings.copy()
        channel_peers = self.lnworker.channel_peers()
        node_info: Optional["NodeInfo"] = None

        while node_stats:
            # randomly pick nodes weighted by node_rating
            pk = choices(list(node_stats.keys()), weights=list(node_ratings.values()), k=1)[0]
            # remove the pk so it doesn't get tried again
            node_stats.pop(pk); node_ratings.pop(pk)
            # node should have compatible features
            node_info = self.network.channel_db.get_node_infos().get(pk, None)
            peer_features = LnFeatures(node_info.features)
            try:
                ln_compare_features(self.lnworker.features, peer_features)
            except IncompatibleLightningFeatures as e:
                self.logger.info("suggested node is incompatible")
                continue

            # don't want to connect to nodes we are already connected to
            if pk in channel_peers:
                continue
            # don't want to connect to nodes we already have a channel with on another device
            if self.lnworker.has_conflicting_backup_with(pk):
                continue
            # node should be on clearnet and have an address saved
            for (hostname, _, _) in self.lnworker.channel_db.get_node_addresses(node_id=pk):
                if not hostname.endswith(".onion"):
                    break
            else:
                continue
            break
        else:
            self.logger.info(f"no suitable channel peer found")
            return None

        alias = node_info.alias if node_info else 'unknown node alias'
        self.logger.info(
            f"node rating for {alias}:\n"
            f"{pformat(self._node_stats[pk])} (score {self._node_ratings[pk]})")

        return pk

    def suggest_peer(self) -> Optional[bytes]:
        """Suggests a LN node to open a channel with.
        Returns a node ID (pubkey).
        """
        self.maybe_analyze_graph()
        if self._node_ratings:
            return self.suggest_node_channel_open()
        else:
            return None
